from fastmcp import FastMCP
from fastmcp.server.auth.providers.workos import AuthKitProvider
from fastmcp.prompts.prompt import Message
from fastmcp.resources import ResourceTemplate

from mqtt_mcp.mqtt_client import AsyncMQTTClient
from mqtt_mcp.settings import Settings


settings = Settings()


async def receive_message(
    topic: str,
    host: str = settings.mqtt.host,
    port: int = settings.mqtt.port,
    username: str | None = settings.mqtt.username,
    password: str | None = settings.mqtt.password,
    timeout: int = 60,
) -> str:
    """Receives a message published to the specified topic, if any."""
    try:
        async with AsyncMQTTClient(host, port, username, password) as client:
            return await client.receive(topic, timeout)
    except Exception as e:
        raise RuntimeError(f"{e}") from e


async def publish_message(
    topic: str,
    message: str,
    host: str = settings.mqtt.host,
    port: int = settings.mqtt.port,
    username: str | None = settings.mqtt.username,
    password: str | None = settings.mqtt.password,
) -> str:
    """Publishes a message to the specified topic."""
    try:
        async with AsyncMQTTClient(host, port, username, password) as client:
            await client.publish(topic, message)
        return f"Publish to {topic} on {host}:{port} has succedeed"
    except Exception as e:
        raise RuntimeError(f"{e}") from e


def mqtt_help() -> list[Message]:
    """Provides examples of how to use the MQTT MCP server."""
    return [
        Message("Here are examples of how to publish and receives messages:"),
        Message('Publish {"foo":"bar"} to topic "devices/foo" on 127.0.0.1:1883.'),
        Message(
            'Receive a message from topic "devices/bar", waiting up to 30 seconds.'
        ),
    ]


def mqtt_error(error: str | None = None) -> list[Message]:
    """Asks the user how to handle an error."""
    return (
        [
            Message(f"ERROR: {error!r}"),
            Message("Would you like to retry, change parameters, or abort?"),
        ]
        if error
        else []
    )


class MQTTMCP(FastMCP):
    def __init__(self, **kwargs):
        super().__init__(
            name="MQTT MCP Server",
            auth=(
                AuthKitProvider(
                    authkit_domain=settings.auth.domain, base_url=settings.auth.url
                )
                if settings.auth.domain and settings.auth.url
                else None
            ),
            **kwargs,
        )

        self.add_template(
            ResourceTemplate.from_function(
                fn=receive_message, uri_template="mqtt://{host}:{port}/{topic*}"
            )
        )

        self.tool(
            receive_message,
            annotations={
                "title": "Receive Message",
                "readOnlyHint": True,
                "openWorldHint": True,
            },
        )

        self.tool(
            publish_message,
            annotations={
                "title": "Publish Message",
                "readOnlyHint": False,
                "openWorldHint": True,
            },
        )

        self.prompt(mqtt_error, name="mqtt_error", tags={"mqtt", "error"})
        self.prompt(mqtt_help, name="mqtt_help", tags={"mqtt", "help"})



import asyncio
import pytest

from fastmcp import Client


@pytest.mark.asyncio
async def test_receive_message(server, mcp):
    """Test receive_message."""
    topic = "foo"
    message = '{"bar":123}'
    async with Client(mcp) as client:

        async def pub():
            await asyncio.sleep(1.0)
            return await client.call_tool(
                "publish_message",
                {
                    "topic": topic,
                    "message": message,
                    "host": server.host,
                    "port": server.port,
                },
            )

        sub = None
        async with asyncio.TaskGroup() as tg:
            sub = tg.create_task(
                client.call_tool(
                    "receive_message",
                    {
                        "topic": topic,
                        "host": server.host,
                        "port": server.port,
                        "timeout": 3,
                    },
                )
            )
            tg.create_task(pub())

        result = sub.result()
        assert len(result.content) == 1
        assert result.content[0].text == message


@pytest.mark.asyncio
async def test_publish_message(server, mcp):
    """Test publish_message."""
    async with Client(mcp) as client:
        result = await client.call_tool(
            "publish_message",
            {
                "topic": "foo",
                "message": '{"bar":456}',
                "host": server.host,
                "port": server.port,
            },
        )
        assert len(result.content) == 1
        assert "succedeed" in result.content[0].text


@pytest.mark.asyncio
async def test_help_prompt(mcp):
    """Test help prompt."""
    async with Client(mcp) as client:
        result = await client.get_prompt("mqtt_help", {})
        assert len(result.messages) == 3


@pytest.mark.asyncio
async def test_error_prompt(mcp):
    """Test error prompt."""
    async with Client(mcp) as client:
        result = await client.get_prompt("mqtt_error", {"error": "Could not read data"})
        assert len(result.messages) == 2



import asyncio
import typer

from mqtt_mcp.server import MQTTMCP


app = typer.Typer(
    name="mqtt-mcp",
    help="MQTTMCP CLI",
)


@app.command()
def run():
    server = MQTTMCP()
    asyncio.run(server.run_async(transport="http"))



import asyncio

from agents import Agent, Runner
from agents.mcp import MCPServer, MCPServerStreamableHttp
from agents.model_settings import ModelSettings


async def run(mcp_server: MCPServer):
    agent = Agent(
        name="Assistant",
        instructions="Use the tools to answer the questions.",
        mcp_servers=[mcp_server],
        model_settings=ModelSettings(tool_choice="required"),
    )

    for prompt in [
        'Publish {"foo":"bar"} to topic "devices/foo" on 127.0.0.1:1883.',
        'Receive a message from topic "devices/bar", waiting up to 30 seconds.',
    ]:
        print(f"Running: {prompt}")
        result = await Runner.run(starting_agent=agent, input=prompt)
        print(result.final_output)


async def main():
    async with MCPServerStreamableHttp(
        name="Streamable HTTP Server",
        params={
            "url": "http://localhost:8000/mcp/",
        },
    ) as server:
        await run(server)


if __name__ == "__main__":
    asyncio.run(main())



"""Async MQTT Client."""

import asyncio

import paho.mqtt.client as mqtt

from typing import Optional


class AsyncioHelper:
    """Integrate paho-mqtt socket callbacks with asyncio event loop."""

    def __init__(self, loop, client):
        self.loop = loop
        self.client = client
        self.client.on_socket_open = self.on_socket_open
        self.client.on_socket_close = self.on_socket_close
        self.client.on_socket_register_write = self.on_socket_register_write
        self.client.on_socket_unregister_write = self.on_socket_unregister_write

    def on_socket_open(self, client, userdata, sock):
        def callback():
            client.loop_read()

        self.loop.add_reader(sock, callback)
        self.misc = self.loop.create_task(self.misc_loop())

    def on_socket_close(self, client, userdata, sock):
        self.loop.remove_reader(sock)
        self.misc.cancel()

    def on_socket_register_write(self, client, userdata, sock):
        def callback():
            client.loop_write()

        self.loop.add_writer(sock, callback)

    def on_socket_unregister_write(self, client, userdata, sock):
        self.loop.remove_writer(sock)

    async def misc_loop(self):
        while self.client.loop_misc() == mqtt.MQTT_ERR_SUCCESS:
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                break


class AsyncMQTTClient:
    """Async MQTT client wrapper."""

    def __init__(
        self,
        host: str,
        port: int = 1883,
        username: str | None = None,
        password: str | None = None,
    ):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2, clean_session=True)
        if self.username and self.password:
            self.client.username_pw_set(self.username, self.password)
        self.future: Optional[asyncio.Future[str]] = None

    async def __aenter__(self) -> "AsyncMQTTClient":
        loop = asyncio.get_running_loop()
        self.helper = AsyncioHelper(loop, self.client)
        self.client.connect(self.host, self.port)
        return self

    async def __aexit__(self, exc_type, exc, tb):
        self.client.disconnect()
        await asyncio.sleep(0)

    async def receive(self, topic: str, timeout: int = 60, qos: int = 1) -> str:
        loop = asyncio.get_running_loop()
        self.future = loop.create_future()

        @self.client.connect_callback()
        def on_connect(client, userdata, flags, reason_code, properties):
            client.subscribe(topic, qos=qos)

        @self.client.message_callback()
        def on_message(client, userdata, message):
            if self.future and not self.future.done():
                self.future.set_result(message.payload.decode())

        try:
            return await asyncio.wait_for(self.future, timeout)
        finally:
            self.future = None

    async def publish(self, topic: str, message: str, qos: int = 1) -> None:
        result = self.client.publish(topic, message, qos=qos)
        if result.rc != mqtt.MQTT_ERR_SUCCESS:
            raise RuntimeError(f"Publish failed with code {result.rc}")
